import re

from friday.spam_handler import log_spam
from .constants import status_not_implemented, status_ok, status_service_unavailable, \
    text_content_types
from .parse import parse
from .version import __version__

methods = ['GET', 'POST', 'get', 'post']


class Response:
    def __init__(self):
        self.content = ''
        self.content_type = 'text/html'
        self.status = status_ok
        self.headers = []

    def __call__(self, start_response):

        if self.headers:
            start_response(str(self.status) + " ", self.headers)
        else:
            self.headers = [('Content-Type', self.content_type), ('Content-Length', str(len(self.content)))]
            start_response(str(self.status) + " ", self.headers)
        # yield content.encode('utf8')

        if self.content_type in text_content_types.values():

            if isinstance(self.content, bytes):
                return [(self.content.decode('utf8', 'ignore')).encode('utf8')]
            else:
                return [self.content.encode('utf8')]

        return [self.content]


class Request:
    def __init__(self, environ=None, path=None, method=None):
        self.environ = environ
        self.method = method
        self.path = path

        self.ip_address = None
        try:
            self.ip_address = environ['REMOTE_ADDR']
        except:
            pass

        self.remote_host = None
        try:
            self.remote_host = environ['REMOTE_HOST']
        except:
            pass

        self.referer = None
        try:
            self.referer = environ['HTTP_REFERER']
        except:
            pass

        self.user_agent = None
        try:
            self.user_agent = environ['HTTP_USER_AGENT']
        except:
            pass

    def info(self):
        return dict(
            path=self.path if self.path is not None else '',
            method=self.method if self.method is not None else '',
            ip_address=self.ip_address if self.ip_address is not None else '',
            remote_host=self.remote_host if self.remote_host is not None else '',
            referer=self.referer if self.referer is not None else '',
            user_agent=self.user_agent if self.user_agent is not None else ''
        )

    def get_range_header(self):

        """ Returns request Range start and end if specified.

        If Range header is not specified returns (None, None)

        """

        range_header = self.environ.get("HTTP_RANGE")

        if range_header is None:
            return (None, None)

        if not range_header.startswith("bytes="):
            # "Not implemented: parsing header Range: %s" % range_header
            return (None, None)

        regex = re.compile(r"^bytes=(\d+)\-(\d+)?")

        rangething = regex.search(range_header)

        if rangething:
            from_val = int(rangething.group(1))
            if rangething.group(2) is not None:
                return (from_val, int(rangething.group(2)))

            else:
                return (from_val, None)

        else:
            print('CANNOT PARSE RANGE HEADER:', range_header)
            return (None, None)


class Friday(object):
    '''
    The Friday object implements a WSGI application
    '''

    def __init__(self, default_response=None, block_ips=None, spam_log_path=None):
        self.spam_log_path = spam_log_path
        self.path = ""
        self.method = None
        self.environ = None
        self.routes = dict()
        self.request = None
        if default_response:
            self.routes['default_response'] = default_response
        self.block_ips = block_ips
        self.response = Response()

        # add default welcome page
        @self.route('/', methods=['GET'])
        def welcome(request, response):
            self.response.content = 'Welcome to Friday Web Framework (' + __version__ + ') for Python by Tennyson Chingombe'
            self.response.status = status_ok

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ, start_response):
        self.method = environ.get('REQUEST_METHOD')
        self.environ = environ

        path = str(environ.get('PATH_INFO')).strip()
        self.path = path if path.endswith('/') else path + '/'

        self.request = Request(environ=self.environ, path=self.path, method=self.method)

        self.response = Response()

        # ip address blocker
        ip_address = self.request.info()['ip_address']
        if self.block_ips:
            if ip_address in self.block_ips:
                print('blocked ip', ip_address)
                self.response.status = status_service_unavailable
                self.response.content = 'Your IP Address is blocked'
                return self.response(start_response)

        if self.method in methods:

            # check if request path has handler in the routes dictionary
            handler, kwargs = self.find_handler(request_path=self.path, method=self.method)

            if handler and kwargs:
                # call handling function

                handler(self.request, self.response, **kwargs)
            elif handler and kwargs is None:
                handler(self.request, self.response)

            else:
                # check if a default response was provided
                if self.method == 'GET' and 'default_response' in self.routes.keys():
                    self.routes['default_response'](self.response)
                else:

                    self.response.status = status_not_implemented
                    self.response.content = 'Not Implemented'
        else:
            self.response.content = "Not implemented"
            self.response.status = status_not_implemented

        return self.response(start_response)

    def find_handler(self, request_path, method=None):
        try:
            handler_dict = self.routes[request_path]
            if method in handler_dict['methods']:
                return handler_dict['function'], None
            else:
                self.response.status = status_not_implemented

                # log spammers here
                if method != 'GET':
                    if self.spam_log_path:
                        log_spam(self.spam_log_path, self.request)
                return None, None
        except Exception as e:
            pass

        for route, handler_dict in self.routes.items():
            parse_result = parse(route, request_path)
            if parse_result:
                # check if the result dictionary is empty or not
                if parse_result.named:
                    # check if request method is supported
                    if method in handler_dict['methods']:
                        return handler_dict['function'], parse_result.named
                    else:
                        self.response.status = status_not_implemented
                        # log spammers here
                        if method != 'GET':
                            if self.spam_log_path:
                                log_spam(self.spam_log_path, self.request)
                        return None, None
                else:
                    return None, None

        return None, None

    def route(self, path=None, methods=['GET']):
        '''
        Function for adding routes to the route dict. Used as a decorator
        :param path:
        :return:
        '''

        path = path.strip()
        if not path.endswith('/'):
            path = path + '/'

        def wrapper(handler):
            self.routes[path] = {'function': handler, 'methods': methods}
            return handler

        return wrapper

    def version(self):
        return __version__
